@@ -70,7 +70,7 @@ atombios_output_dac_setup(xf86OutputPtr output, DisplayModePtr mode)
     DAC_ENCODER_CONTROL_PS_ALLOCATION disp_data;
     AtomBiosArgRec data;
     unsigned char *space;
-    int index, num = 0;
+    int index = 0, num = 0;
 
     if (radeon_encoder == NULL)
 	return ATOM_NOT_IMPLEMENTED;
@@ -264,7 +264,7 @@ atombios_output_digital_setup(xf86OutputPtr output, DisplayModePtr mode)
     LVDS_ENCODER_CONTROL_PS_ALLOCATION_V2 disp_data2;
     AtomBiosArgRec data;
     unsigned char *space;
-    int index;
+    int index = 0;
     int major, minor;
     int lvds_misc = 0;
 
@@ -430,7 +430,10 @@ atombios_get_encoder_mode(xf86OutputPtr output)
 	return ATOM_ENCODER_MODE_LVDS;
 	break;
     case CONNECTOR_DISPLAY_PORT:
-	return ATOM_ENCODER_MODE_DP;
+	if (radeon_output->MonType == MT_DP)
+	    return ATOM_ENCODER_MODE_DP;
+	else
+	    return atombios_maybe_hdmi_mode(output);
 	break;
     case CONNECTOR_DVI_A:
     case CONNECTOR_VGA:
@@ -448,6 +451,42 @@ atombios_get_encoder_mode(xf86OutputPtr output)
 
 }
 
+static const int dp_clocks[] = {
+    16200,
+    27000,
+    32400,
+    54000,
+    0,
+    0,
+    64800,
+    108000,
+};
+static const int num_dp_clocks = sizeof(dp_clocks) / sizeof(int);
+
+static int
+dp_lanes_for_mode_clock(int mode_clock)
+{
+    int i;
+    
+    for (i = 0; i < num_dp_clocks; i++)
+	if (dp_clocks[i] > (mode_clock / 10))
+	    return (i / 2) + 1;
+
+    return 0;
+}
+
+static int
+dp_link_clock_for_mode_clock(int mode_clock)
+{
+    int i;
+
+    for (i = 0; i < num_dp_clocks; i++)
+	if (dp_clocks[i] > (mode_clock / 10))
+	    return (dp_clocks[i % 2]);
+
+    return 0;
+}
+
 static int
 atombios_output_dig_encoder_setup(xf86OutputPtr output, DisplayModePtr mode)
 {
@@ -458,7 +497,7 @@ atombios_output_dig_encoder_setup(xf86OutputPtr output, DisplayModePtr mode)
     DIG_ENCODER_CONTROL_PS_ALLOCATION disp_data;
     AtomBiosArgRec data;
     unsigned char *space;
-    int index, major, minor, num = 0;
+    int index = 0, major, minor, num = 0;
 
     if (radeon_encoder == NULL)
 	return ATOM_NOT_IMPLEMENTED;
@@ -507,9 +546,11 @@ atombios_output_dig_encoder_setup(xf86OutputPtr output, DisplayModePtr mode)
 	}
     } else {
 	switch (radeon_encoder->encoder_id) {
+	case ENCODER_OBJECT_ID_INTERNAL_UNIPHY:
+	    disp_data.ucConfig = ATOM_ENCODER_CONFIG_UNIPHY;
+	    break;
 	case ENCODER_OBJECT_ID_INTERNAL_TMDS1:
 	case ENCODER_OBJECT_ID_INTERNAL_KLDSCP_TMDS1:
-	case ENCODER_OBJECT_ID_INTERNAL_UNIPHY:
 	    disp_data.ucConfig = ATOM_ENCODER_CONFIG_TRANSMITTER1;
 	    break;
 	case ENCODER_OBJECT_ID_INTERNAL_LVDS:
@@ -520,7 +561,19 @@ atombios_output_dig_encoder_setup(xf86OutputPtr output, DisplayModePtr mode)
 	}
     }
 
-    if (mode->Clock > 165000) {
+    disp_data.ucEncoderMode = atombios_get_encoder_mode(output);
+
+    if (disp_data.ucEncoderMode == ATOM_ENCODER_MODE_DP) {
+	if (radeon_output->linkb)
+	    disp_data.ucConfig |= ATOM_ENCODER_CONFIG_LINKB;
+	else
+	    disp_data.ucConfig |= ATOM_ENCODER_CONFIG_LINKA;
+
+	if (dp_link_clock_for_mode_clock(mode->Clock) == 27000)
+	    disp_data.ucConfig |= ATOM_ENCODER_CONFIG_DPLINKRATE_2_70GHZ;
+
+	disp_data.ucLaneNum = dp_lanes_for_mode_clock(mode->Clock);
+    } else if (mode->Clock > 165000) {
 	disp_data.ucConfig |= ATOM_ENCODER_CONFIG_LINKA_B;
 	disp_data.ucLaneNum = 8;
     } else {
@@ -528,11 +581,10 @@ atombios_output_dig_encoder_setup(xf86OutputPtr output, DisplayModePtr mode)
 	    disp_data.ucConfig |= ATOM_ENCODER_CONFIG_LINKB;
 	else
 	    disp_data.ucConfig |= ATOM_ENCODER_CONFIG_LINKA;
+
 	disp_data.ucLaneNum = 4;
     }
 
-    disp_data.ucEncoderMode = atombios_get_encoder_mode(output);
-
     data.exec.index = index;
     data.exec.dataSpace = (void *)&space;
     data.exec.pspace = &disp_data;
@@ -562,7 +614,7 @@ atombios_output_dig_transmitter_setup(xf86OutputPtr output, DisplayModePtr mode)
     union dig_transmitter_control disp_data;
     AtomBiosArgRec data;
     unsigned char *space;
-    int index, num = 0;
+    int index = 0, num = 0;
     int major, minor;
 
     if (radeon_encoder == NULL)
@@ -592,7 +644,11 @@ atombios_output_dig_transmitter_setup(xf86OutputPtr output, DisplayModePtr mode)
     disp_data.v1.ucAction = ATOM_TRANSMITTER_ACTION_ENABLE;
 
     if (IS_DCE32_VARIANT) {
-	if (mode->Clock > 165000) {
+	if (radeon_output->MonType == MT_DP) {
+	    disp_data.v2.usPixelClock =
+		cpu_to_le16(dp_link_clock_for_mode_clock(mode->Clock));
+	    disp_data.v2.acConfig.fDPConnector = 1;
+	} else if (mode->Clock > 165000) {
 	    disp_data.v2.usPixelClock = cpu_to_le16((mode->Clock * 10 * 2) / 100);
 	    disp_data.v2.acConfig.fDualLinkConnector = 1;
 	} else {
@@ -625,12 +681,27 @@ atombios_output_dig_transmitter_setup(xf86OutputPtr output, DisplayModePtr mode)
 	}
     } else {
 	disp_data.v1.ucConfig = ATOM_TRANSMITTER_CONFIG_CLKSRC_PPLL;
-	disp_data.v1.usPixelClock = cpu_to_le16((mode->Clock) / 10);
+
+	if (radeon_output->MonType == MT_DP)
+	    disp_data.v1.usPixelClock =
+		cpu_to_le16(dp_link_clock_for_mode_clock(mode->Clock));
+	else
+	    disp_data.v1.usPixelClock = cpu_to_le16((mode->Clock) / 10);
 
 	switch (radeon_encoder->encoder_id) {
+	case ENCODER_OBJECT_ID_INTERNAL_UNIPHY:
+	    if (radeon_output->MonType == MT_DP) {
+		if (radeon_output->linkb)
+		    disp_data.v1.ucConfig |= ATOM_TRANSMITTER_CONFIG_LINKB
+					| ATOM_TRANSMITTER_CONFIG_DIG2_ENCODER;
+		else
+		    disp_data.v1.ucConfig |= ATOM_TRANSMITTER_CONFIG_LINKA
+					| ATOM_TRANSMITTER_CONFIG_DIG1_ENCODER;
+		break;
+	    };
+	    /* else, fall through */
 	case ENCODER_OBJECT_ID_INTERNAL_TMDS1:
 	case ENCODER_OBJECT_ID_INTERNAL_KLDSCP_TMDS1:
-	case ENCODER_OBJECT_ID_INTERNAL_UNIPHY:
 	    disp_data.v1.ucConfig |= ATOM_TRANSMITTER_CONFIG_DIG1_ENCODER;
 	    if (info->IsIGP) {
 		if (mode->Clock > 165000) {
@@ -685,11 +756,15 @@ atombios_output_dig_transmitter_setup(xf86OutputPtr output, DisplayModePtr mode)
 	}
 
 	if (radeon_output->active_device & (ATOM_DEVICE_DFP_SUPPORT)) {
-	    if (radeon_output->coherent_mode) {
+	    if (radeon_output->coherent_mode &&
+		radeon_output->MonType != MT_DP) {
 		disp_data.v1.ucConfig |= ATOM_TRANSMITTER_CONFIG_COHERENT;
-		xf86DrvMsg(output->scrn->scrnIndex, X_INFO, "DIG%d transmitter: Coherent Mode enabled\n", num);
-	    } else
-		xf86DrvMsg(output->scrn->scrnIndex, X_INFO, "DIG%d transmitter: Coherent Mode disabled\n", num);
+		xf86DrvMsg(output->scrn->scrnIndex, X_INFO,
+			"DIG%d transmitter: Coherent Mode enabled\n", num);
+	    } else {
+		xf86DrvMsg(output->scrn->scrnIndex, X_INFO,
+			"DIG%d transmitter: Coherent Mode disabled\n", num);
+	    }
 	}
     }
     radeon_output->transmitter_config = disp_data.v1.ucConfig;
@@ -994,6 +1069,61 @@ atombios_output_yuv_setup(xf86OutputPtr output, Bool enable)
 
 }
 
+static int
+atombios_output_overscan_setup(xf86OutputPtr output, DisplayModePtr mode, DisplayModePtr adjusted_mode)
+{
+    RADEONOutputPrivatePtr radeon_output = output->driver_private;
+    RADEONCrtcPrivatePtr radeon_crtc = output->crtc->driver_private;
+    RADEONInfoPtr info       = RADEONPTR(output->scrn);
+    SET_CRTC_OVERSCAN_PS_ALLOCATION overscan_param;
+    AtomBiosArgRec data;
+    unsigned char *space;
+    memset(&overscan_param, 0, sizeof(overscan_param));
+
+    overscan_param.usOverscanRight = 0;
+    overscan_param.usOverscanLeft = 0;
+    overscan_param.usOverscanBottom = 0;
+    overscan_param.usOverscanTop = 0;
+    overscan_param.ucCRTC = radeon_crtc->crtc_id;
+
+    if (radeon_output->Flags & RADEON_USE_RMX) {
+	if (radeon_output->rmx_type == RMX_FULL) {
+	    overscan_param.usOverscanRight = 0;
+	    overscan_param.usOverscanLeft = 0;
+	    overscan_param.usOverscanBottom = 0;
+	    overscan_param.usOverscanTop = 0;
+	} else if (radeon_output->rmx_type == RMX_CENTER) {
+	    overscan_param.usOverscanTop = (adjusted_mode->CrtcVDisplay - mode->CrtcVDisplay) / 2;
+	    overscan_param.usOverscanBottom = (adjusted_mode->CrtcVDisplay - mode->CrtcVDisplay) / 2;
+	    overscan_param.usOverscanLeft = (adjusted_mode->CrtcHDisplay - mode->CrtcHDisplay) / 2;
+	    overscan_param.usOverscanRight = (adjusted_mode->CrtcHDisplay - mode->CrtcHDisplay) / 2;
+	} else if (radeon_output->rmx_type == RMX_ASPECT) {
+	    int a1 = mode->CrtcVDisplay * adjusted_mode->CrtcHDisplay;
+	    int a2 = adjusted_mode->CrtcVDisplay * mode->CrtcHDisplay;
+
+	    if (a1 > a2) {
+		overscan_param.usOverscanLeft = (adjusted_mode->CrtcHDisplay - (a2 / mode->CrtcVDisplay)) / 2;
+		overscan_param.usOverscanRight = (adjusted_mode->CrtcHDisplay - (a2 / mode->CrtcVDisplay)) / 2;
+	    } else if (a2 > a1) {
+		overscan_param.usOverscanLeft = (adjusted_mode->CrtcVDisplay - (a1 / mode->CrtcHDisplay)) / 2;
+		overscan_param.usOverscanRight = (adjusted_mode->CrtcVDisplay - (a1 / mode->CrtcHDisplay)) / 2;
+	    }
+	}
+    }
+
+    data.exec.index = GetIndexIntoMasterTable(COMMAND, SetCRTC_OverScan);
+    data.exec.dataSpace = (void *)&space;
+    data.exec.pspace = &overscan_param;
+
+    if (RHDAtomBiosFunc(info->atomBIOS->scrnIndex, info->atomBIOS, ATOMBIOS_EXEC, &data) == ATOM_SUCCESS) {
+	ErrorF("Set CRTC %d Overscan success\n", radeon_crtc->crtc_id);
+	return ATOM_SUCCESS ;
+    }
+
+    ErrorF("Set CRTC %d Overscan failed\n", radeon_crtc->crtc_id);
+    return ATOM_NOT_IMPLEMENTED;
+}
+
 static int
 atombios_output_scaler_setup(xf86OutputPtr output, DisplayModePtr mode)
 {
@@ -1005,6 +1135,9 @@ atombios_output_scaler_setup(xf86OutputPtr output, DisplayModePtr mode)
     AtomBiosArgRec data;
     unsigned char *space;
 
+    if (!IS_AVIVO_VARIANT && radeon_crtc->crtc_id)
+	return ATOM_SUCCESS;
+
     memset(&disp_data, 0, sizeof(disp_data));
 
     disp_data.ucScaler = radeon_crtc->crtc_id;
@@ -1051,9 +1184,14 @@ atombios_output_scaler_setup(xf86OutputPtr output, DisplayModePtr mode)
 	    disp_data.ucEnable = ATOM_SCALER_EXPANSION;
 	else if (radeon_output->rmx_type == RMX_CENTER)
 	    disp_data.ucEnable = ATOM_SCALER_CENTER;
+	else if (radeon_output->rmx_type == RMX_ASPECT)
+	    disp_data.ucEnable = ATOM_SCALER_EXPANSION;
     } else {
 	ErrorF("Not using RMX\n");
-	disp_data.ucEnable = ATOM_SCALER_DISABLE;
+	if (IS_AVIVO_VARIANT)
+	    disp_data.ucEnable = ATOM_SCALER_DISABLE;
+	else
+	    disp_data.ucEnable = ATOM_SCALER_CENTER;
     }
 
     data.exec.index = GetIndexIntoMasterTable(COMMAND, EnableScaler);
@@ -1266,8 +1404,47 @@ atombios_set_output_crtc_source(xf86OutputPtr output)
 	case 0:
 	case 1:
 	default:
-	    crtc_src_param.ucCRTC = radeon_crtc->crtc_id;
-	    crtc_src_param.ucDevice = radeon_get_device_index(radeon_output->active_device);
+	    if (IS_AVIVO_VARIANT)
+		crtc_src_param.ucCRTC = radeon_crtc->crtc_id;
+	    else
+		crtc_src_param.ucCRTC = radeon_crtc->crtc_id << 2;
+	    switch (radeon_encoder->encoder_id) {
+	    case ENCODER_OBJECT_ID_INTERNAL_TMDS1:
+	    case ENCODER_OBJECT_ID_INTERNAL_KLDSCP_TMDS1:
+		crtc_src_param.ucDevice = ATOM_DEVICE_DFP1_INDEX;
+		break;
+	    case ENCODER_OBJECT_ID_INTERNAL_LVDS:
+	    case ENCODER_OBJECT_ID_INTERNAL_LVTM1:
+	    case ENCODER_OBJECT_ID_INTERNAL_KLDSCP_LVTMA:
+		if (radeon_output->active_device & ATOM_DEVICE_LCD1_SUPPORT)
+		    crtc_src_param.ucDevice = ATOM_DEVICE_LCD1_INDEX;
+		else
+		    crtc_src_param.ucDevice = ATOM_DEVICE_DFP3_INDEX;
+		break;
+	    case ENCODER_OBJECT_ID_INTERNAL_DVO1:
+	    case ENCODER_OBJECT_ID_INTERNAL_DDI:
+	    case ENCODER_OBJECT_ID_INTERNAL_KLDSCP_DVO1:
+		crtc_src_param.ucDevice = ATOM_DEVICE_DFP2_INDEX;
+		break;
+	    case ENCODER_OBJECT_ID_INTERNAL_DAC1:
+	    case ENCODER_OBJECT_ID_INTERNAL_KLDSCP_DAC1:
+		if (radeon_output->active_device & (ATOM_DEVICE_TV_SUPPORT))
+		    crtc_src_param.ucDevice = ATOM_DEVICE_TV1_INDEX;
+		else if (radeon_output->active_device & (ATOM_DEVICE_CV_SUPPORT))
+		    crtc_src_param.ucDevice = ATOM_DEVICE_CV_INDEX;
+		else
+		    crtc_src_param.ucDevice = ATOM_DEVICE_CRT1_INDEX;
+		break;
+	    case ENCODER_OBJECT_ID_INTERNAL_DAC2:
+	    case ENCODER_OBJECT_ID_INTERNAL_KLDSCP_DAC2:
+		if (radeon_output->active_device & (ATOM_DEVICE_TV_SUPPORT))
+		    crtc_src_param.ucDevice = ATOM_DEVICE_TV1_INDEX;
+		else if (radeon_output->active_device & (ATOM_DEVICE_CV_SUPPORT))
+		    crtc_src_param.ucDevice = ATOM_DEVICE_CV_INDEX;
+		else
+		    crtc_src_param.ucDevice = ATOM_DEVICE_CRT2_INDEX;
+		break;
+	    }
 	    data.exec.pspace = &crtc_src_param;
 	    /*ErrorF("device sourced: 0x%x\n", crtc_src_param.ucDevice);*/
 	    break;
@@ -1372,7 +1549,7 @@ atombios_apply_output_quirks(xf86OutputPtr output, DisplayModePtr mode)
     }
 
     /* set scaler clears this on some chips */
-    if (mode->Flags & V_INTERLACE)
+    if (IS_AVIVO_VARIANT && (mode->Flags & V_INTERLACE))
 	OUTREG(AVIVO_D1MODE_DATA_FORMAT + radeon_crtc->crtc_offset, AVIVO_D1MODE_INTERLEAVE_EN);
 }
 
@@ -1383,16 +1560,20 @@ atombios_output_mode_set(xf86OutputPtr output,
 {
     RADEONOutputPrivatePtr radeon_output = output->driver_private;
     radeon_encoder_ptr radeon_encoder = radeon_get_encoder(output);
-
+    RADEONInfoPtr info       = RADEONPTR(output->scrn);
     if (radeon_encoder == NULL)
         return;
 
-    atombios_output_scaler_setup(output, mode);
+    atombios_output_overscan_setup(output, mode, adjusted_mode);
+    atombios_output_scaler_setup(output, adjusted_mode);
     atombios_set_output_crtc_source(output);
-    if (radeon_output->active_device & (ATOM_DEVICE_CV_SUPPORT | ATOM_DEVICE_TV_SUPPORT))
-	atombios_output_yuv_setup(output, TRUE);
-    else
-	atombios_output_yuv_setup(output, FALSE);
+
+    if (IS_AVIVO_VARIANT) {
+	if (radeon_output->active_device & (ATOM_DEVICE_CV_SUPPORT | ATOM_DEVICE_TV_SUPPORT))
+	    atombios_output_yuv_setup(output, TRUE);
+	else
+	    atombios_output_yuv_setup(output, FALSE);
+    }
 
     switch (radeon_encoder->encoder_id) {
     case ENCODER_OBJECT_ID_INTERNAL_TMDS1:
