@@ -78,6 +78,7 @@
 				/* Driver data structures */
 #include "radeon.h"
 #include "radeon_reg.h"
+#include "r600_reg.h"
 #include "radeon_macros.h"
 #include "radeon_probe.h"
 #include "radeon_version.h"
@@ -92,6 +93,7 @@
 				/* X and server generic header files */
 #include "xf86.h"
 
+static void R600EngineReset(ScrnInfoPtr pScrn);
 
 #ifdef USE_XAA
 static struct {
@@ -149,6 +151,37 @@ void RADEONWaitForFifoFunction(ScrnInfoPtr pScrn, int entries)
     }
 }
 
+void R600WaitForFifoFunction(ScrnInfoPtr pScrn, int entries)
+{
+    RADEONInfoPtr  info       = RADEONPTR(pScrn);
+    unsigned char *RADEONMMIO = info->MMIO;
+    int            i;
+
+    for (;;) {
+	for (i = 0; i < RADEON_TIMEOUT; i++) {
+	    if (info->ChipFamily >= CHIP_FAMILY_RV770)
+		info->accel_state->fifo_slots =
+		    INREG(R600_GRBM_STATUS) & R700_CMDFIFO_AVAIL_MASK;
+	    else
+		info->accel_state->fifo_slots =
+		    INREG(R600_GRBM_STATUS) & R600_CMDFIFO_AVAIL_MASK;
+	    if (info->accel_state->fifo_slots >= entries) return;
+	}
+	xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, RADEON_LOGLEVEL_DEBUG,
+		       "FIFO timed out: stat=0x%08x\n",
+		       (unsigned int)INREG(R600_GRBM_STATUS));
+	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+		   "FIFO timed out, resetting engine...\n");
+	R600EngineReset(pScrn);
+#ifdef XF86DRI
+	if (info->directRenderingEnabled) {
+	    RADEONCP_RESET(pScrn, info);
+	    RADEONCP_START(pScrn, info);
+	}
+#endif
+    }
+}
+
 /* Flush all dirty data in the Pixel Cache to memory */
 void RADEONEngineFlush(ScrnInfoPtr pScrn)
 {
@@ -156,9 +189,6 @@ void RADEONEngineFlush(ScrnInfoPtr pScrn)
     unsigned char *RADEONMMIO = info->MMIO;
     int            i;
 
-    if (info->ChipFamily >= CHIP_FAMILY_R600)
-        return;
-
     if (info->ChipFamily <= CHIP_FAMILY_RV280) {
 	OUTREGP(RADEON_RB3D_DSTCACHE_CTLSTAT,
 		RADEON_RB3D_DC_FLUSH_ALL,
@@ -198,8 +228,6 @@ void RADEONEngineReset(ScrnInfoPtr pScrn)
     uint32_t       rbbm_soft_reset;
     uint32_t       host_path_cntl;
 
-    if (info->ChipFamily >= CHIP_FAMILY_R600)
-        return;
     /* The following RBBM_SOFT_RESET sequence can help un-wedge
      * an R300 after the command processor got stuck.
      */
@@ -310,6 +338,35 @@ void RADEONEngineReset(ScrnInfoPtr pScrn)
     OUTPLL(pScrn, RADEON_MCLK_CNTL, mclk_cntl);
 }
 
+/* Reset graphics card to known state */
+static void R600EngineReset(ScrnInfoPtr pScrn)
+{
+    RADEONInfoPtr  info       = RADEONPTR(pScrn);
+    unsigned char *RADEONMMIO = info->MMIO;
+    uint32_t cp_ptr, cp_me_cntl, cp_rb_cntl;
+
+    cp_ptr = INREG(R600_CP_RB_WPTR);
+
+    cp_me_cntl = INREG(R600_CP_ME_CNTL);
+    OUTREG(R600_CP_ME_CNTL, 0x10000000);
+
+    OUTREG(R600_GRBM_SOFT_RESET, 0x7fff);
+    INREG(R600_GRBM_SOFT_RESET);
+    usleep (50);
+    OUTREG(R600_GRBM_SOFT_RESET, 0);
+    INREG(R600_GRBM_SOFT_RESET);
+
+    OUTREG(R600_CP_RB_WPTR_DELAY, 0);
+    cp_rb_cntl = INREG(R600_CP_RB_CNTL);
+    OUTREG(R600_CP_RB_CNTL, 0x80000000);
+
+    OUTREG(R600_CP_RB_RPTR_WR, cp_ptr);
+    OUTREG(R600_CP_RB_WPTR, cp_ptr);
+    OUTREG(R600_CP_RB_CNTL, cp_rb_cntl);
+    OUTREG(R600_CP_ME_CNTL, cp_me_cntl);
+
+}
+
 /* Restore the acceleration hardware to its previous state */
 void RADEONEngineRestore(ScrnInfoPtr pScrn)
 {
@@ -611,8 +668,12 @@ drmBufPtr RADEONCPGetBuffer(ScrnInfoPtr pScrn)
 
 	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
 		   "GetBuffer timed out, resetting engine...\n");
-	RADEONEngineReset(pScrn);
-	RADEONEngineRestore(pScrn);
+
+	if (info->ChipFamily < CHIP_FAMILY_R600) {
+	    RADEONEngineReset(pScrn);
+	    RADEONEngineRestore(pScrn);
+	} else
+	    R600EngineReset(pScrn);
 
 	/* Always restart the engine when doing CP 2D acceleration */
 	RADEONCP_RESET(pScrn, info);
@@ -627,6 +688,8 @@ void RADEONCPFlushIndirect(ScrnInfoPtr pScrn, int discard)
     drmBufPtr          buffer = info->cp->indirectBuffer;
     int                start  = info->cp->indirectStart;
     drm_radeon_indirect_t  indirect;
+    RING_LOCALS;
+    RADEONCP_REFRESH(pScrn, info);
 
     if (!buffer) return;
     if (start == buffer->used && !discard) return;
@@ -636,6 +699,14 @@ void RADEONCPFlushIndirect(ScrnInfoPtr pScrn, int discard)
 		   buffer->idx);
     }
 
+    if (info->ChipFamily >= CHIP_FAMILY_R600) {
+	while (buffer->used & 0x3c){
+	    BEGIN_RING(1);
+	    OUT_RING(CP_PACKET2()); /* fill up to multiple of 16 dwords */
+	    ADVANCE_RING();
+	}
+    }
+
     indirect.idx     = buffer->idx;
     indirect.start   = start;
     indirect.end     = buffer->used;
@@ -664,6 +735,19 @@ void RADEONCPReleaseIndirect(ScrnInfoPtr pScrn)
     drmBufPtr          buffer = info->cp->indirectBuffer;
     int                start  = info->cp->indirectStart;
     drm_radeon_indirect_t  indirect;
+    RING_LOCALS;
+    RADEONCP_REFRESH(pScrn, info);
+
+
+    if (info->ChipFamily >= CHIP_FAMILY_R600) {
+	if (buffer) {
+	    while (buffer->used & 0x3c) {
+		BEGIN_RING(1);
+		OUT_RING(CP_PACKET2()); /* fill up to multiple of 16 dwords */
+		ADVANCE_RING();
+	    }
+	}
+    }
 
     info->cp->indirectBuffer = NULL;
     info->cp->indirectStart  = 0;
@@ -926,20 +1010,26 @@ Bool RADEONAccelInit(ScreenPtr pScreen)
     ScrnInfoPtr    pScrn = xf86Screens[pScreen->myNum];
     RADEONInfoPtr  info  = RADEONPTR(pScrn);
 
-    if (info->ChipFamily >= CHIP_FAMILY_R600)
-	return FALSE;
-
 #ifdef USE_EXA
     if (info->useEXA) {
 # ifdef XF86DRI
 	if (info->directRenderingEnabled) {
-	    if (!RADEONDrawInitCP(pScreen))
-		return FALSE;
+	    if (info->ChipFamily >= CHIP_FAMILY_R600) {
+		if (!R600DrawInit(pScreen))
+		    return FALSE;
+	    } else {
+		if (!RADEONDrawInitCP(pScreen))
+		    return FALSE;
+	    }
 	} else
 # endif /* XF86DRI */
 	{
-	    if (!RADEONDrawInitMMIO(pScreen))
+	    if (info->ChipFamily >= CHIP_FAMILY_R600)
 		return FALSE;
+	    else {
+		if (!RADEONDrawInitMMIO(pScreen))
+		    return FALSE;
+	    }
 	}
     }
 #endif /* USE_EXA */
@@ -947,6 +1037,9 @@ Bool RADEONAccelInit(ScreenPtr pScreen)
     if (!info->useEXA) {
 	XAAInfoRecPtr  a;
 
+	if (info->ChipFamily >= CHIP_FAMILY_R600)
+	    return FALSE;
+
 	if (!(a = info->accel_state->accel = XAACreateInfoRec())) {
 	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "XAACreateInfoRec Error\n");
 	    return FALSE;
